"""
Contains all of the elements in the periodic table and dictionaries that
store all of the data found in the periodic table for that element, indexed
by the element's symbol. For consistency with AMBER, a fictitious element
'EP' is added to all of the arrays that is just an Extra Point, with no mass
or any other meaningful attribute. It's just a container to put an extra
charge
"""
# Data descriptions:
#
#  KNOWN_ELEMENTS: number of known elements
#  Element:        array whose indices are the atomic number corresponding to
#                  the element
#  AtomicNum:      dictionary matching chemical symbol to atomic number
#  Mass:           dictionary matching chemical symbol to atomic mass
#  Name:           dicionary matching chemical symbol to their full name
#  OriginName:     dictionary matching chemical symbol to the name from which
#                  the symbol comes
#  Phase:          Lists standard phase that the element appears in

KNOWN_ELEMENTS = 118

AtomicNum = { 'H'  :  1, 'He' :  2, 'Li' :  3, 'Be' :  4, 'B'  :  5, 'C'  :  6,
              'N'  :  7, 'O'  :  8, 'F'  :  9, 'Ne' : 10, 'Na' : 11, 'Mg' : 12,
              'Al' : 13, 'Si' : 14, 'P'  : 15, 'S'  : 16, 'Cl' : 17, 'Ar' : 18,
              'K'  : 19, 'Ca' : 20, 'Sc' : 21, 'Ti' : 22, 'V'  : 23, 'Cr' : 24,
              'Mn' : 25, 'Fe' : 26, 'Co' : 27, 'Ni' : 28, 'Cu' : 29, 'Zn' : 30,
              'Ga' : 31, 'Ge' : 32, 'As' : 33, 'Se' : 34, 'Br' : 35, 'Kr' : 36,
              'Rb' : 37, 'Sr' : 38, 'Y'  : 39, 'Zr' : 40, 'Nb' : 41, 'Mo' : 42,
              'Tc' : 43, 'Ru' : 44, 'Rh' : 45, 'Pd' : 46, 'Ag' : 47, 'Cd' : 48,
              'In' : 49, 'Sn' : 50, 'Sb' : 51, 'Te' : 52, 'I'  : 53, 'Xe' : 54,
              'Cs' : 55, 'Ba' : 56, 'La' : 57, 'Ce' : 58, 'Pr' : 59, 'Nd' : 60,
              'Pm' : 61, 'Sm' : 62, 'Eu' : 63, 'Gd' : 64, 'Tb' : 65, 'Dy' : 66,
              'Ho' : 67, 'Er' : 68, 'Tm' : 69, 'Yb' : 70, 'Lu' : 71, 'Hf' : 72,
              'Ta' : 73, 'W'  : 74, 'Re' : 75, 'Os' : 76, 'Ir' : 77, 'Pt' : 78,
              'Au' : 79, 'Hg' : 80, 'Tl' : 81, 'Pb' : 82, 'Bi' : 83, 'Po' : 84,
              'At' : 85, 'Rn' : 86, 'Fr' : 87, 'Ra' : 88, 'Ac' : 89, 'Th' : 90,
              'Pa' : 91, 'U'  : 92, 'Np' : 93, 'Pu' : 94, 'Am' : 95, 'Cm' : 96,
              'Bk' : 97, 'Cf' : 98, 'Es' : 99, 'Fm' :100, 'Md' :101, 'No' :102,
              'Lr' :103, 'Rf' :104, 'Db' :105, 'Sg' :106, 'Bh' :107, 'Hs' :108,
              'Mt' :109, 'Ds' :110, 'Rg' :111, 'Cn' :112, 'Nh' :113, 'Fl' :114,
              'Mc' :115, 'Lv' :116, 'Ts' :117, 'Og' :118, 'EP' : 0 , 'LP' :  0,
              'Lp' :  0, 'Ep' :  0}

Element = [ 'EP',
            'H' ,'He','Li','Be','B' ,'C' ,'N' ,'O' ,'F' ,'Ne','Na','Mg',
            'Al','Si','P' ,'S' ,'Cl','Ar','K' ,'Ca','Sc','Ti','V' ,'Cr',
            'Mn','Fe','Co','Ni','Cu','Zn','Ga','Ge','As','Se','Br','Kr',
            'Rb','Sr','Y' ,'Zr','Nb','Mo','Tc','Ru','Rh','Pd','Ag','Cd',
            'In','Sn','Sb','Te','I' ,'Xe','Cs','Ba','La','Ce','Pr','Nd',
            'Pm','Sm','Eu','Gd','Tb','Dy','Ho','Er','Tm','Yb','Lu','Hf',
            'Ta','W' ,'Re','Os','Ir','Pt','Au','Hg','Tl','Pb','Bi','Po',
            'At','Rn','Fr','Ra','Ac','Th','Pa','U' ,'Np','Pu','Am','Cm',
            'Bk','Cf','Es','Fm','Md','No','Lr','Rf','Db','Sg','Bh','Hs',
            'Mt','Ds','Rg','Cn','Nh','Fl','Mc','Lv','Ts','Og' ]

Mass = { 'H'  :   1.0079 , 'He' :   4.0026 , 'Li' :   6.941  ,
         'Be' :   9.0122 , 'B'  :  10.811  , 'C'  :  12.0107 ,
         'N'  :  14.0067 , 'O'  :  15.9994 , 'F'  :  18.9984 ,
         'Ne' :  20.1797 , 'Na' :  22.9898 , 'Mg' :  24.3050 ,
         'Al' :  26.9815 , 'Si' :  28.0855 , 'P'  :  30.9738 ,
         'S'  :  32.065  , 'Cl' :  35.453  , 'Ar' :  39.948  ,
         'K'  :  39.0983 , 'Ca' :  40.078  , 'Sc' :  44.9559 ,
         'Ti' :  47.867  , 'V'  :  50.9415 , 'Cr' :  51.9961 ,
         'Mn' :  54.9380 , 'Fe' :  55.845  , 'Co' :  58.9331 ,
         'Ni' :  58.6934 , 'Cu' :  63.546  , 'Zn' :  65.409  ,
         'Ga' :  69.723  , 'Ge' :  72.64   , 'As' :  74.9216 ,
         'Se' :  78.96   , 'Br' :  79.904  , 'Kr' :  83.798  ,
         'Rb' :  85.4678 , 'Sr' :  87.62   , 'Y'  :  88.9059 ,
         'Zr' :  91.224  , 'Nb' :  92.9064 , 'Mo' :  95.94   ,
         'Tc' :  98.     , 'Ru' : 101.07   , 'Rh' : 102.9055 ,
         'Pd' : 106.42   , 'Ag' : 107.8682 , 'Cd' : 112.411  ,
         'In' : 114.818  , 'Sn' : 118.710  , 'Sb' : 121.760  ,
         'Te' : 127.60   , 'I'  : 126.9045 , 'Xe' : 131.293  ,
         'Cs' : 132.9055 , 'Ba' : 137.327  , 'La' : 138.9055 ,
         'Ce' : 140.116  , 'Pr' : 140.9077 , 'Nd' : 144.242  ,
         'Pm' : 145.     , 'Sm' : 150.36   , 'Eu' : 151.964  ,
         'Gd' : 157.25   , 'Tb' : 158.9254 , 'Dy' : 162.500  ,
         'Ho' : 164.9303 , 'Er' : 167.259  , 'Tm' : 168.9342 ,
         'Yb' : 173.04   , 'Lu' : 174.967  , 'Hf' : 178.49   ,
         'Ta' : 180.9479 , 'W'  : 183.84   , 'Re' : 186.207  ,
         'Os' : 190.23   , 'Ir' : 192.217  , 'Pt' : 195.084  ,
         'Au' : 196.9666 , 'Hg' : 200.59   , 'Tl' : 204.3833 ,
         'Pb' : 207.2    , 'Bi' : 208.9804 , 'Po' : 209.     ,
         'At' : 210.     , 'Rn' : 222.     , 'Fr' : 223.     ,
         'Ra' : 226.     , 'Ac' : 227.     , 'Th' : 232.0381 ,
         'Pa' : 231.0359 , 'U'  : 238.0289 , 'Np' : 237.     ,
         'Pu' : 244.     , 'Am' : 243.     , 'Cm' : 247.     ,
         'Bk' : 247.     , 'Cf' : 251.     , 'Es' : 252.     ,
         'Fm' : 257.     , 'Md' : 258.     , 'No' : 259.     ,
         'Lr' : 262.     , 'Rf' : 261.     , 'Db' : 262.     ,
         'Sg' : 266.     , 'Bh' : 264.     , 'Hs' : 277.     ,
         'Mt' : 268.     , 'Ds' : 281.     , 'Rg' : 272.     ,
         'Cn' : 285.     , 'Nh' : 286.     , 'Fl' : 289.     ,
         'Mc' : 289.     , 'Lv' : 293.     , 'Ts' : 294.     ,
         'Og' : 294.     , 'EP' : 0.000000 }

Name = { 'H'  : 'Hydrogen'     ,'He' : 'Helium'       ,'Li' : 'Lithium'      ,
         'Be' : 'Beryllium'    ,'B'  : 'Boron'        ,'C'  : 'Carbon'       ,
         'N'  : 'Nitrogen'     ,'O'  : 'Oxygen'       ,'F'  : 'Fluorine'     ,
         'Ne' : 'Neon'         ,'Na' : 'Sodium'       ,'Mg' : 'Magnesium'    ,
         'Al' : 'Aluminum'     ,'Si' : 'Silicon'      ,'P'  : 'Phosphorus'   ,
         'S'  : 'Sulfur'       ,'Cl' : 'Chlorine'     ,'Ar' : 'Argon'        ,
         'K'  : 'Potassium'    ,'Ca' : 'Calcium'      ,'Sc' : 'Scandium'     ,
         'Ti' : 'Titanium'     ,'V'  : 'Vanadium'     ,'Cr' : 'Chromium'     ,
         'Mn' : 'Manganese'    ,'Fe' : 'Iron'         ,'Co' : 'Cobalt'       ,
         'Ni' : 'Nickel'       ,'Cu' : 'Copper'       ,'Zn' : 'Zinc'         ,
         'Ga' : 'Gallium'      ,'Ge' : 'Germanium'    ,'As' : 'Arsenic'      ,
         'Se' : 'Selenium'     ,'Br' : 'Bromine'      ,'Kr' : 'Krypton'      ,
         'Rb' : 'Rubidium'     ,'Sr' : 'Strontium'    ,'Y'  : 'Yttrium'      ,
         'Zr' : 'Zirconium'    ,'Nb' : 'Niobium'      ,'Mo' : 'Molybdenum'   ,
         'Tc' : 'Technetium'   ,'Ru' : 'Ruthenium'    ,'Rh' : 'Rhodium'      ,
         'Pd' : 'Palladium'    ,'Ag' : 'Silver'       ,'Cd' : 'Cadmium'      ,
         'In' : 'Indium'       ,'Sn' : 'Tin'          ,'Sb' : 'Antimony'     ,
         'Te' : 'Tellurium'    ,'I'  : 'Iodine'       ,'Xe' : 'Xenon'        ,
         'Cs' : 'Cesium'       ,'Ba' : 'Barium'       ,'La' : 'Lanthanum'    ,
         'Ce' : 'Cerium'       ,'Pr' : 'Praseodymium' ,'Nd' : 'Neodymium'    ,
         'Pm' : 'Promethium'   ,'Sm' : 'Samarium'     ,'Eu' : 'Europium'     ,
         'Gd' : 'Gadolinium'   ,'Tb' : 'Terbium'      ,'Dy' : 'Dysprosium'   ,
         'Ho' : 'Holmium'      ,'Er' : 'Erbium'       ,'Tm' : 'Thulium'      ,
         'Yb' : 'Ytterbium'    ,'Lu' : 'Lutetium'     ,'Hf' : 'Hafnium'      ,
         'Ta' : 'Tantalum'     ,'W'  : 'Tungsten'     ,'Re' : 'Rhenium'      ,
         'Os' : 'Osmium'       ,'Ir' : 'Iridium'      ,'Pt' : 'Platinum'     ,
         'Au' : 'Gold'         ,'Hg' : 'Mercury'      ,'Tl' : 'Thallium'     ,
         'Pb' : 'Lead'         ,'Bi' : 'Bismuth'      ,'Po' : 'Polonium'     ,
         'At' : 'Astatine'     ,'Rn' : 'Radon'        ,'Fr' : 'Francium'     ,
         'Ra' : 'Radium'       ,'Ac' : 'Actinium'     ,'Th' : 'Thorium'      ,
         'Pa' : 'Proactinium'  ,'U'  : 'Uranium'      ,'Np' : 'Neptunium'    ,
         'Pu' : 'Plutonium'    ,'Am' : 'Americium'    ,'Cm' : 'Curium'       ,
         'Bk' : 'Berkelium'    ,'Cf' : 'Californium'  ,'Es' : 'Einsteinium'  ,
         'Fm' : 'Fermium'      ,'Md' : 'Mendelevium'  ,'No' : 'Nobelium'     ,
         'Lr' : 'Lawrencium'   ,'Rf' : 'Rutherfordium','Db' : 'Dubnium'      ,
         'Sg' : 'Seaborgium'   ,'Bh' : 'Bohrium'      ,'Hs' : 'Hassium'      ,
         'Mt' : 'Meitnerium'   ,'Ds' : 'Darmstadtium' ,'Rg' : 'Roentgenium'  ,
         'Cn' : 'Copernicium'  ,'Nh' : 'Nihonium'     ,'Fl' : 'Flerovium'    ,
         'Mc' : 'Moscovium'    ,'Lv' : 'Livermorium'  ,'Ts' : 'Tennessine'   ,
         'Og' : 'Oganesson'    ,'EP' : 'Extra Point'  ,'LP' : 'Extra Point'  ,
         'Ep' : 'Extra Point'  ,'Lp' : 'Extra Point'}

OriginName = {
        'H'  : 'Hydrogen'   ,'He' : 'Helium'       ,'Li' : 'Lithium'      ,
        'Be' : 'Beryllium'  ,'B'  : 'Boron'        ,'C'  : 'Carbon'       ,
        'N'  : 'Nitrogen'   ,'O'  : 'Oxygen'       ,'F'  : 'Fluorine'     ,
        'Ne' : 'Neon'       ,'Na' : 'Natrium'      ,'Mg' : 'Magnesium'    ,
        'Al' : 'Aluminum'   ,'Si' : 'Silicon'      ,'P'  : 'Phosphorus'   ,
        'S'  : 'Sulfur'     ,'Cl' : 'Chlorine'     ,'Ar' : 'Argon'        ,
        'K'  : 'Kalium'     ,'Ca' : 'Calcium'      ,'Sc' : 'Scandium'     ,
        'Ti' : 'Titanium'   ,'V'  : 'Vanadium'     ,'Cr' : 'Chromium'     ,
        'Mn' : 'Manganese'  ,'Fe' : 'Ferrum'       ,'Co' : 'Cobalt'       ,
        'Ni' : 'Nickel'     ,'Cu' : 'Cuprum'       ,'Zn' : 'Zinc'         ,
        'Ga' : 'Gallium'    ,'Ge' : 'Germanium'    ,'As' : 'Arsenic'      ,
        'Se' : 'Selenium'   ,'Br' : 'Bromine'      ,'Kr' : 'Krypton'      ,
        'Rb' : 'Rubidium'   ,'Sr' : 'Strontium'    ,'Y'  : 'Yttrium'      ,
        'Zr' : 'Zirconium'  ,'Nb' : 'Niobium'      ,'Mo' : 'Molybdenum'   ,
        'Tc' : 'Technetium' ,'Ru' : 'Ruthenium'    ,'Rh' : 'Rhodium'      ,
        'Pd' : 'Palladium'  ,'Ag' : 'Argentum'     ,'Cd' : 'Cadmium'      ,
        'In' : 'Indium'     ,'Sn' : 'Stannum'      ,'Sb' : 'Stibium'      ,
        'Te' : 'Tellurium'  ,'I'  : 'Iodine'       ,'Xe' : 'Xenon'        ,
        'Cs' : 'Cesium'     ,'Ba' : 'Barium'       ,'La' : 'Lanthanum'    ,
        'Ce' : 'Cerium'     ,'Pr' : 'Praseodymium' ,'Nd' : 'Neodymium'    ,
        'Pm' : 'Promethium' ,'Sm' : 'Samarium'     ,'Eu' : 'Europium'     ,
        'Gd' : 'Gadolinium' ,'Tb' : 'Terbium'      ,'Dy' : 'Dysprosium'   ,
        'Ho' : 'Holmium'    ,'Er' : 'Erbium'       ,'Tm' : 'Thulium'      ,
        'Yb' : 'Ytterbium'  ,'Lu' : 'Lutetium'     ,'Hf' : 'Hafnium'      ,
        'Ta' : 'Tantalum'   ,'W'  : 'Wolfram'      ,'Re' : 'Rhenium'      ,
        'Os' : 'Osmium'     ,'Ir' : 'Iridium'      ,'Pt' : 'Platinum'     ,
        'Au' : 'Aurum'      ,'Hg' : 'Hydrargyrum'  ,'Tl' : 'Thallium'     ,
        'Pb' : 'Plumbum'    ,'Bi' : 'Bismuth'      ,'Po' : 'Polonium'     ,
        'At' : 'Astatine'   ,'Rn' : 'Radon'        ,'Fr' : 'Francium'     ,
        'Ra' : 'Radium'     ,'Ac' : 'Actinium'     ,'Th' : 'Thorium'      ,
        'Pa' : 'Proactinium','U'  : 'Uranium'      ,'Np' : 'Neptunium'    ,
        'Pu' : 'Plutonium'  ,'Am' : 'Americium'    ,'Cm' : 'Curium'       ,
        'Bk' : 'Berkelium'  ,'Cf' : 'Californium'  ,'Es' : 'Einsteinium'  ,
        'Fm' : 'Fermium'    ,'Md' : 'Mendelevium'  ,'No' : 'Nobelium'     ,
        'Lr' : 'Lawrencium' ,'Rf' : 'Rutherfordium','Db' : 'Dubnium'      ,
        'Sg' : 'Seaborgium' ,'Bh' : 'Bohrium'      ,'Hs' : 'Hassium'      ,
        'Mt' : 'Meitnerium' ,'Ds' : 'Darmstadtium' ,'Rg' : 'Roentgenium'  ,
        'Cn' : 'Copernicium','Nh' : 'Nihonium'     ,'Fl' : 'Flerovium'    ,
        'Mc' : 'Moscovium'  ,'Lv' : 'Livermorium'  ,'Ts' : 'Tennessine'   ,
        'Og' : 'Oganesson'  ,'EP' : 'Extra Point'  ,'LP' : 'Extra Point'  ,
        'Ep' : 'Extra Point','Lp' : 'Extra Point'
}

Phase = { 'H'  : 'Gas'          ,'He' : 'Gas'          ,'Li' : 'Solid'        ,
          'Be' : 'Solid'        ,'B'  : 'Solid'        ,'C'  : 'Solid'        ,
          'N'  : 'Gas'          ,'O'  : 'Gas'          ,'F'  : 'Gas'          ,
          'Ne' : 'Gas'          ,'Na' : 'Solid'        ,'Mg' : 'Solid'        ,
          'Al' : 'Solid'        ,'Si' : 'Solid'        ,'P'  : 'Solid'        ,
          'S'  : 'Solid'        ,'Cl' : 'Gas'          ,'Ar' : 'Gas'          ,
          'K'  : 'Solid'        ,'Ca' : 'Solid'        ,'Sc' : 'Solid'        ,
          'Ti' : 'Solid'        ,'V'  : 'Solid'        ,'Cr' : 'Solid'        ,
          'Mn' : 'Solid'        ,'Fe' : 'Solid'        ,'Co' : 'Solid'        ,
          'Ni' : 'Solid'        ,'Cu' : 'Solid'        ,'Zn' : 'Solid'        ,
          'Ga' : 'Solid'        ,'Ge' : 'Solid'        ,'As' : 'Solid'        ,
          'Se' : 'Solid'        ,'Br' : 'Liquid'       ,'Kr' : 'Gas'          ,
          'Rb' : 'Solid'        ,'Sr' : 'Solid'        ,'Y'  : 'Solid'        ,
          'Zr' : 'Solid'        ,'Nb' : 'Solid'        ,'Mo' : 'Solid'        ,
          'Tc' : 'Solid'        ,'Ru' : 'Solid'        ,'Rh' : 'Solid'        ,
          'Pd' : 'Solid'        ,'Ag' : 'Solid'        ,'Cd' : 'Solid'        ,
          'In' : 'Solid'        ,'Sn' : 'Solid'        ,'Sb' : 'Solid'        ,
          'Te' : 'Solid'        ,'I'  : 'Solid'        ,'Xe' : 'Gas'          ,
          'Cs' : 'Solid'        ,'Ba' : 'Solid'        ,'La' : 'Solid'        ,
          'Ce' : 'Solid'        ,'Pr' : 'Solid'        ,'Nd' : 'Solid'        ,
          'Pm' : 'Solid'        ,'Sm' : 'Solid'        ,'Eu' : 'Solid'        ,
          'Gd' : 'Solid'        ,'Tb' : 'Solid'        ,'Dy' : 'Solid'        ,
          'Ho' : 'Solid'        ,'Er' : 'Solid'        ,'Tm' : 'Solid'        ,
          'Yb' : 'Solid'        ,'Lu' : 'Solid'        ,'Hf' : 'Solid'        ,
          'Ta' : 'Solid'        ,'W'  : 'Solid'        ,'Re' : 'Solid'        ,
          'Os' : 'Solid'        ,'Ir' : 'Solid'        ,'Pt' : 'Solid'        ,
          'Au' : 'Solid'        ,'Hg' : 'Liquid'       ,'Tl' : 'Solid'        ,
          'Pb' : 'Solid'        ,'Bi' : 'Solid'        ,'Po' : 'Solid'        ,
          'At' : 'Solid'        ,'Rn' : 'Solid'        ,'Fr' : 'Solid'        ,
          'Ra' : 'Gas'          ,'Ac' : 'Solid'        ,'Th' : 'Solid'        ,
          'Pa' : 'Solid'        ,'U'  : 'Solid'        ,'Np' : 'Solid'        ,
          'Pu' : 'Solid'        ,'Am' : 'Solid'        ,'Cm' : 'Solid'        ,
          'Bk' : 'Solid'        ,'Cf' : 'Solid'        ,'Es' : 'Solid'        ,
          'Fm' : 'Solid'        ,'Md' : 'Solid'        ,'No' : 'Solid'        ,
          'Lr' : 'Solid'        ,'Rf' : 'Unknown'      ,'Db' : 'Unknown'      ,
          'Sg' : 'Unknown'      ,'Bh' : 'Unknown'      ,'Hs' : 'Unknown'      ,
          'Mt' : 'Unknown'      ,'Ds' : 'Unknown'      ,'Rg' : 'Unknown'      ,
          'Cn' : 'Unknown'      ,'Nh' : 'Unknown'      ,'Fl' : 'Unknown'      ,
          'Mc' : 'Unknown'      ,'Lv' : 'Unknown'      ,'Ts' : 'Unknown'      ,
          'Og' : 'Unknown'      ,'EP' : 'N/A'          ,'Ep' : 'N/A'          ,
          'LP' : 'N/A'          ,'Lp' : 'N/A'
}

_sorted_masses = sorted(Mass.items(), key=lambda x: x[1])

def element_by_mass(mass):
    """
    Determine the element that has a mass closest to the input mass

    Parameters
    ----------
    mass : float
        The atomic mass to compare to

    Returns
    -------
    element: str
        The returned string is the name of the element whose atomic mass is
        closest to the input mass

    Notes
    -----
    This actually fails (i.e., produces poor predictions) for some cases when
    masses have changed -- particularly in the case of Hydrogen mass
    repartitioning.
    """
    diff = mass + 1
    best_guess = 'EP'

    for element, element_mass in _sorted_masses:
        d = abs(element_mass - mass)
        if d < diff:
            best_guess = element
            diff = d
        else:
            break

    return best_guess

def element_by_name(name):
    """
    Determine the element based on the name of an atom. This is very naive. It
    first tries to match the first letter of the element. If that doesn't work,
    it tries to match the first *two* letters. If that still doesn't work, it
    defaults to an extra point

    Parameters
    ----------
    name : str
        Name of the atom to determine an element for

    Returns
    -------
    element : str
        The name of the best-matching element

    Notes
    -----
    This should be a last-case scenario for guessing element information. For
    instance, Ca will never be matched, since calcium atoms will be tagged as
    carbon before the second letter is tried. This is usually OK for
    biomolecules, but you are better off using the mass or, even better, an
    appropriate representation of the atomic number to begin with
    """
    name = name.strip()
    if len(name) == 0:
        return Element[0]
    try:
        atomic_number = AtomicNum[name[0].upper()]
    except KeyError:
        sym = name[:2]
        try:
            sym = '%s%s' % (sym[0].upper(), sym[1].lower())
            atomic_number = AtomicNum[sym]
        except (KeyError, IndexError):
            atomic_number = 0 # give up

    return Element[atomic_number]

# Add some mass aliases here. We need to do it *after* _sorted_masses is created above, since
# the element_by_mass routine which uses the _sorted_masses assumes that the masses are all
# monotonically strictly increasing
Mass.update(dict(Ep=0.0, LP=0.0, Lp=0.0))



"""
This module contains parsers for sdf file format 
extension described at https://www.cas.org/content/chemical-suppliers/example-sdf
"""
from __future__ import print_function, division, absolute_import
import linecache

from parmed.formats.registry import FileFormatType
from parmed.utils.six import add_metaclass
from parmed import rdkit

@add_metaclass(FileFormatType)
class SDFFile(object):
    """ Class to read SDF file """

    @staticmethod
    def id_format(filename):
        """ Identify the file as a SDF file format or not

        Parameters
        ----------
        filename : str
            Name of the file to test whether or not it is a sdf file

        Returns
        -------
        is_fmt : bool
            True if it is a sdf file, False otherwise
        """
        words = linecache.getline(filename, lineno=4).split()
        return len(words) >= 3 and words[-1] in ('V2000', 'V3000')

    @staticmethod
    def parse(filename, structure=False):
        return rdkit.from_sdf(filename, structure=structure)



"""
This module contains classes for parsing and processing Amber parameter files.

Author: Jason M. Swails
Contributors:
Date: Aug. 11, 2015
"""
from __future__ import division, print_function

from collections import defaultdict
from contextlib import closing
import math
import os
import re
import warnings
from .offlib import AmberOFFLibrary
from ..constants import TINY
from ..exceptions import ParameterError, AmberWarning, ParameterWarning
from ..modeller.residue import ResidueTemplateContainer
from ..formats.mol2 import Mol2File
from ..formats.registry import FileFormatType
from ..parameters import ParameterSet
from ..periodic_table import Mass, element_by_mass, AtomicNum
from ..topologyobjects import (AtomType, BondType, AngleType, DihedralType,
                                    DihedralTypeList)
from ..utils.io import genopen
from ..utils.six import add_metaclass, string_types, iteritems, PY2
from ..utils.six.moves import map
if PY2:
    from collections import Sequence
else:
    from collections.abc import Sequence

# parameter file regexes
subs = dict(FLOATRE=r'([+-]?(?:\d+(?:\.\d*)?|\.\d+))',
            FILENAMERE=r'''(".+?"|'.+?'|[\S]*)''')
_bondre = re.compile(r'(..?)-(..?)\s+%(FLOATRE)s\s+%(FLOATRE)s' % subs)
_anglere = re.compile(r'(..?)-(..?)-(..?)\s+%(FLOATRE)s\s+%(FLOATRE)s' % subs)
_dihedre = re.compile(r'(..?)-(..?)-(..?)-(..?)\s+%(FLOATRE)s\s+'
                      '%(FLOATRE)s\s+%(FLOATRE)s\s+%(FLOATRE)s' % subs)
_dihed2re = re.compile(r'\s*%(FLOATRE)s\s+%(FLOATRE)s\s+%(FLOATRE)s\s+'
                       '%(FLOATRE)s' % subs)
_sceere = re.compile(r'SCEE=\s*%(FLOATRE)s' % subs)
_scnbre = re.compile(r'SCNB=\s*%(FLOATRE)s' % subs)
_impropre = re.compile(r'(..?)-(..?)-(..?)-(..?)\s+'
                       '%(FLOATRE)s\s+%(FLOATRE)s\s+%(FLOATRE)s' % subs)
# Leaprc regexes
_atomtypere = re.compile(r"""({\s*["']([\w\+\-]+)["']\s*["'](\w+)["']\s*"""
                         r"""["'](\w+)["']\s*})""")
_loadparamsre = re.compile(r'loadamberparams\s+%(FILENAMERE)s' % subs, re.I)
_loadoffre = re.compile(r'loadoff\s+%(FILENAMERE)s' % subs, re.I)
_loadmol2re = re.compile(r'(\S+)\s*=\s*loadmol[23]\s*%(FILENAMERE)s' % subs, re.I)
del subs

def _find_amber_file(fname, search_oldff):
    """
    Finds an Amber file. Looks in the current directory, then the following
    locations:

    - $AMBERHOME/dat/leap/lib
    - $AMBERHOME/dat/leap/parm
    """
    from parmed.amber import AMBERHOME
    if len(fname) > 1 and {fname[0], fname[-1]} in ({'"'}, {"'"}):
        # Strip quotes
        fname = fname[1:-1]
    leapdir = os.path.join(AMBERHOME, 'dat', 'leap')
    paths = [os.getcwd(), os.path.join(leapdir, 'lib'), os.path.join(leapdir, 'parm')]
    if search_oldff:
        paths.append(os.path.join(leapdir, 'lib', 'oldff'))
    for path in paths:
        if os.path.exists(os.path.join(path, fname)):
            return os.path.join(path, fname)
    raise FileNotFoundError('Cannot find Amber file [%s] in paths %s' % (fname, paths))

@add_metaclass(FileFormatType)
class AmberParameterSet(ParameterSet):
    """ Class storing parameters from an Amber parameter set

    Parameters
    ----------
    filenames : str, list of str, file-like, or list of file-like; optional
        Either the name of a file or a list of filenames from which parameters
        should be parsed.

    Notes
    -----
    Order is important in the list of files provided. The parameters are loaded
    in the order they are provided, and any parameters that are specified in
    multiple places are overwritten (that is, the *last* occurrence is the
    parameter type that is used)

    See Also
    --------
    :class:`parmed.parameters.ParameterSet`
    """

    #===================================================

    @staticmethod
    def id_format(filename):
        """
        Identifies the file type as either an Amber-style frcmod or parm.dat
        file.

        Parameters
        ----------
        filename : str
            Name of the file to check format for

        Returns
        -------
        is_fmt : bool
            True if it is an Amber-style parameter file. False otherwise.
        """
        with closing(genopen(filename, 'r')) as f:
            f.readline()
            line = f.readline()
            if not line.strip(): # Must be an frcmod file
                while line and not line.strip():
                    line = f.readline()
                if not line:
                    return False
                if line.rstrip() not in ('MASS', 'BOND', 'ANGLE', 'ANGL', 'DIHE', 'DIHED',
                                         'DIHEDRAL', 'IMPR', 'IMPROP', 'IMPROPER', 'NONB', 'NONBON',
                                         'NONBOND', 'NONBONDED'):
                    return False
            if line.rstrip() in ('MASS', 'BOND', 'ANGLE', 'ANGL', 'DIHE', 'DIHED', 'DIHEDRAL',
                                 'IMPR', 'IMPROP', 'IMPROPER', 'NONB', 'NONBON', 'NONBOND',
                                 'NONBONDED'):
                return True # frcmod file
            # This must be an atom definition in the parm.dat file
            words = line.split()
            if len(words) < 2:
                return False
            # The first word is the atom type (must be <= 2 characters, and not
            # an integer)
            if len(words[0]) > 2: return False
            try:
                float(words[0])
                return False
            except ValueError:
                pass
            try:
                float(words[1])
            except ValueError:
                return False
            # Polarizability might not be present...
            try:
                float(words[2])
            except (IndexError, ValueError):
                # UGLY: Check the mass and make sure it matches the element's
                # mass to within 1 amu. Do our best to guess the element. If
                # it's a two-letter element, the element might be a 2-letter
                # element (like Br), or it might be a 1-letter element specified
                # by either the first or second letter. Check all possibilities.
                # Special-case instances like CA, which are just as likely (more
                # likely?) to be a carbon atom as it is to be a calcium (i.e.,
                # check for carbon atoms in anything that starts with C). If at
                # any point, what *should* be the mass doesn't match the mass of
                # the guessed element, tag this as *not* a parameter file
                if len(words[0]) == 2:
                    if words[0][0].isalpha():
                        if words[0][1].isalpha():
                            key = words[0][0].upper() + words[0][1].lower()
                            if key in Mass and abs(Mass[key] - float(words[1])) > 1:
                                if key[0] == 'C' and abs(Mass['C'] - float(words[1])) > 1:
                                    return False
                                elif key[0] != 'C':
                                    return False
                            elif key not in Mass:
                                if key[0] in Mass and abs(Mass[key[0]] - float(words[1])) > 1:
                                    return False
                                else:
                                    return False
                        else:
                            key = words[0][0].upper()
                            if key in Mass and abs(Mass[key] - float(words[1])) > 1:
                                return False
                            elif key not in Mass:
                                return False
                    else:
                        key = words[0][1].upper()
                        if key in Mass:
                            if abs(Mass[key] - float(words[1])) > 1:
                                return False
                        else:
                            return False
                else:
                    key = words[0][0].upper()
                    if key in Mass:
                        if abs(Mass[key] - float(words[1])) > 1:
                            return False
                    else:
                        return False
            if len(words) > 3:
                # Heuristic, but anything that comes after the polarizability is
                # a comment, and I have yet to see a leading comment that is a
                # number
                try:
                    float(words[3])
                    return False
                except ValueError:
                    return True
            else:
                return True

    #===================================================

    def __init__(self, *filenames):
        super(AmberParameterSet, self).__init__()
        self.default_scee = 1.2
        self.default_scnb = 2.0
        self.titles = []
        for filename in filenames:
            if isinstance(filename, string_types):
                if AmberOFFLibrary.id_format(filename):
                    self.residues.update(AmberOFFLibrary.parse(filename))
                else:
                    self.load_parameters(filename)
            elif isinstance(filename, Sequence):
                for fname in filename:
                    if AmberOFFLibrary.id_format(fname):
                        self.residues.update(AmberOFFLibrary.parse(fname))
                    else:
                        self.load_parameters(fname)
            else:
                # Assume open file object
                self.load_parameters(filename)

    #===================================================

    @classmethod
    def from_leaprc(cls, fname, search_oldff=False):
        """ Load a parameter set from a leaprc file

        Parameters
        ----------
        fname : str or file-like
            Name of the file or open file-object from which a leaprc-style file
            will be read

        search_oldff : bool, optional, default=False
            If True, search the oldff directories in the main Amber leap
            folders. Default is False

        Notes
        -----
        This does not read all parts of a leaprc file -- only those pertinent to
        defining force field information. For instance, the following sections
        and commands are processed:

        - addAtomTypes
        - loadAmberParams
        - loadOFF
        - loadMol2
        - loadMol3
        """
        params = cls()
        if isinstance(fname, string_types):
            f = genopen(fname, 'r')
            own_handle = True
        else:
            f = fname
            own_handle = False
        # To make parsing easier, and because leaprc files are usually quite
        # short, I'll read the whole file into memory
        def joinlines(lines):
            newlines = []
            composite = []
            for line in lines:
                if line.endswith('\\\n'):
                    composite.append(line[:-2])
                    continue
                else:
                    composite.append(line)
                newlines.append(''.join(composite))
                composite = []
            if composite:
                newlines.append(''.join(composite))
            return newlines
        lines = joinlines(map(lambda line: line if '#' not in line else line[:line.index('#')], f))
        text = ''.join(lines)
        if own_handle: f.close()
        lowertext = text.lower() # commands are case-insensitive
        # Now process the parameter files
        def process_fname(fname):
            if fname[0] in ('"', "'"):
                fname = fname[1:-1]
            fname = fname.replace('_BSTOKEN_', r'\ ').replace(r'\ ', ' ')
            return fname
        for line in lines:
            line = line.replace(r'\ ', '_BSTOKEN_')
            if _loadparamsre.findall(line):
                fname = process_fname(_loadparamsre.findall(line)[0])
                params.load_parameters(_find_amber_file(fname, search_oldff))
            elif _loadoffre.findall(line):
                fname = process_fname(_loadoffre.findall(line)[0])
                params.residues.update(AmberOFFLibrary.parse(_find_amber_file(fname, search_oldff)))
            elif _loadmol2re.findall(line):
                (resname, fname), = _loadmol2re.findall(line)
                residue = Mol2File.parse(_find_amber_file(fname, search_oldff))
                if isinstance(residue, ResidueTemplateContainer):
                    warnings.warn('Multi-residue mol2 files not supported by tleap. Loading anyway '
                                  'using names in mol2', AmberWarning)
                    for res in residue:
                        params.residues[res.name] = res
                else:
                    params.residues[resname] = residue
        # Now process the addAtomTypes
        try:
            idx = lowertext.index('addatomtypes')
        except ValueError:
            # Does not exist in this file
            atom_types_str = ''
        else:
            i = idx + len('addatomtypes')
            while i < len(text) and text[i] != '{':
                if text[i] not in '\r\n\t ':
                    raise ParameterError('Unsupported addAtomTypes syntax in leaprc file')
                i += 1
            if i == len(text):
                raise ParameterError('Unsupported addAtomTypes syntax in leaprc file')
            # We are at our first brace
            chars = []
            nopen = 1
            i += 1
            while i < len(text):
                char = text[i]
                if char == '{':
                    nopen += 1
                elif char == '}':
                    nopen -= 1
                    if nopen == 0: break
                elif char == '\n':
                    char = ' '
                chars.append(char)
                i += 1
            atom_types_str = ''.join(chars).strip()
        for _, name, symb, hyb in _atomtypere.findall(atom_types_str):
            if symb not in AtomicNum:
                raise ParameterError('%s is not a recognized element' % symb)
            if name in params.atom_types:
                params.atom_types[name].atomic_number = AtomicNum[symb]
        return params

    #===================================================

    @classmethod
    def from_structure(cls, struct):
        """ Extracts known parameters from a Structure instance

        Parameters
        ----------
        struct : :class:`parmed.structure.Structure`
            The parametrized ``Structure`` instance from which to extract
            parameters into a ParameterSet

        Returns
        -------
        params : :class:`ParameterSet`
            The parameter set with all parameters defined in the Structure
        """
        return super(AmberParameterSet, cls).from_structure(struct, allow_unequal_duplicates=False)

    #===================================================

    def load_parameters(self, fname):
        """ Load a set of parameters from a single parameter file

        Parameters
        ----------
        fname : str or file-like
            Parameter file to parse
        """
        if isinstance(fname, string_types):
            f = genopen(fname, 'r')
            own_handle = True
        else:
            f = fname
            own_handle = False
        self.titles.append(f.readline().strip())
        try:
            for line in f:
                if not line.strip():
                    return self._parse_frcmod(f, line)
                elif line.strip() in ('MASS', 'BOND', 'ANGLE', 'ANGL', 'DIHE', 'DIHED', 'DIHEDRAL',
                                      'IMPR', 'IMPROP', 'IMPROPER', 'NONB', 'NONBON', 'NONBOND',
                                      'NONBONDED'):
                    return self._parse_frcmod(f, line)
                else:
                    return self._parse_parm_dat(f, line)
        finally:
            if own_handle:
                f.close()

    #===================================================

    def _parse_frcmod(self, f, line):
        """ Parses an frcmod file from an open file object """
        def fiter():
            yield line
            for l in f: yield l
        section = None
        finished_diheds = defaultdict(lambda: True)
        key = None
        for line in fiter():
            line = line.rstrip()
            if not line: continue
            if line.startswith('MASS'):
                section = 'MASS'
                continue
            elif line.startswith('BOND'):
                section = 'BOND'
                continue
            elif line.startswith('ANGL'):
                section = 'ANGLE'
                continue
            elif line.startswith('DIHE'):
                section = 'DIHEDRAL'
                continue
            elif line.startswith('IMPR'):
                section = 'IMPROPER'
                continue
            elif line.startswith('NONB'):
                section = 'NONBOND'
                continue
            elif line.startswith('LJEDIT'):
                section = 'NBFIX'
                continue

            if section == 'MASS':
                self._process_mass_line(line)
            elif section == 'BOND':
                self._process_bond_line(line)
            elif section == 'ANGLE':
                self._process_angle_line(line)
            elif section == 'DIHEDRAL':
                key = self._process_dihedral_line(line, finished_diheds, key)
            elif section == 'IMPROPER':
                self._process_improper_line(line)
            elif section == 'NONBOND':
                self._process_nonbond_line(line)
            elif section == 'NBFIX':
                self._process_nbfix_line(line)

    #===================================================

    def _parse_parm_dat(self, f, line):
        """ Internal parser for parm.dat files from open file handle """
        def fiter():
            yield line
            for l in f: yield l
            # Keep yielding empty string after file has ended
            yield ''
        fiter = fiter()
        rawline = next(fiter)
        finished_diheds = defaultdict(lambda: True)
        # Parse the masses
        while rawline:
            line = rawline.strip()
            if not line:
                break
            self._process_mass_line(line)
            rawline = next(fiter)
        next(fiter) # Skip the list of hydrophobic atom types
        # Process the bonds
        rawline = next(fiter)
        while rawline:
            line = rawline.strip()
            if not line:
                break
            self._process_bond_line(line)
            rawline = next(fiter)
        # Process the angles
        rawline = next(fiter)
        while rawline:
            line = rawline.strip()
            if not line:
                break
            self._process_angle_line(line)
            rawline = next(fiter)
        # Process the dihedrals
        rawline = next(fiter)
        key = None
        while rawline:
            line = rawline.strip()
            if not line:
                break
            key = self._process_dihedral_line(line, finished_diheds, key)
            rawline = next(fiter)
        # Process the impropers
        rawline = next(fiter)
        while rawline:
            line = rawline.strip()
            if not line:
                break
            self._process_improper_line(line)
            rawline = next(fiter)
        # Process the 10-12 terms
        rawline = next(fiter)
        while rawline:
            line = rawline.strip()
            if not line:
                break
            try:
                a1, a2, acoef, bcoef = line.split()[:4]
                acoef = float(acoef)
                bcoef = float(bcoef)
            except ValueError:
                raise ParameterError('Trouble parsing 10-12 terms')
            if acoef != 0 or bcoef != 0:
                raise ParameterError('10-12 potential not supported in AmberParameterSet currently')
            rawline = next(fiter)
        # Process 12-6 terms. Get Equivalencing first
        rawline = next(fiter)
        equivalent_ljtypes = dict()
        equivalent_types = defaultdict(list)
        while rawline:
            line = rawline.strip()
            if not line:
                break
            words = line.split()
            for typ in words[1:]:
                equivalent_ljtypes[typ] = words[0]
                equivalent_types[words[0]].append(typ)
            rawline = next(fiter)
        words = next(fiter).split()
        if len(words) < 2:
            raise ParameterError('Could not parse the kind of nonbonded '
                                 'parameters in Amber parameter file')
        if words[1].upper() != 'RE':
            raise ParameterError('Only RE nonbonded parameters supported')
        rawline = next(fiter)
        while rawline:
            line = rawline.strip()
            if not line:
                break
            self._process_nonbond_line(line)
            rawline = next(fiter)
        # Now assign all of the equivalenced atoms
        for atyp, otyp in iteritems(equivalent_ljtypes):
            otyp = self.atom_types[otyp]
            if atyp in self.atom_types:
                if (self.atom_types[atyp].rmin is not None and
                        self.atom_types[atyp].epsilon is not None):
                    if (abs(otyp.epsilon-self.atom_types[atyp].epsilon) > TINY
                            or abs(otyp.rmin-self.atom_types[atyp].rmin) > TINY):
                        warnings.warn('Equivalency defined between %s and %s but parameters are '
                                      'not equal' % (otyp.name, atyp), AmberWarning)
                        # Remove from equivalent types
                        equivalent_types[otyp.name].remove(atyp)
                        continue
                self.atom_types[atyp].set_lj_params(otyp.epsilon, otyp.rmin)
        line = next(fiter).strip()
        if line == 'LJEDIT':
            rawline = next(fiter)
            while rawline:
                line = rawline.strip()
                if not line:
                    break
                self._process_nbfix_line(line, equivalent_types)
                rawline = next(fiter)

    #===================================================

    # Private methods for processing parts of the file
    def _process_mass_line(self, line):
        words = line.split()
        try:
            mass = float(words[1])
        except ValueError:
            raise ParameterError('Could not convert mass to float [%s]' % words[1])
        except IndexError:
            raise ParameterError('Error parsing MASS line. Not enough tokens')
        if words[0] in self.atom_types:
            self.atom_types[words[0]].mass = mass
        elif words[0] in ('EP', 'LP'):
            atype = AtomType(words[0], len(self.atom_types)+1, mass, 0)
            self.atom_types[words[0]] = atype
        else:
            atype = AtomType(words[0], len(self.atom_types)+1, mass,
                             AtomicNum[element_by_mass(mass)])
            self.atom_types[words[0]] = atype

    #===================================================

    def _process_bond_line(self, line):
        rematch = _bondre.match(line)
        if not rematch:
            raise ParameterError('Could not understand BOND line [%s]' % line)
        a1, a2, k, eq = rematch.groups()
        a1 = a1.strip(); a2 = a2.strip()
        typ = BondType(float(k), float(eq))
        self.bond_types[(a1, a2)] = typ
        self.bond_types[(a2, a1)] = typ

    def _process_angle_line(self, line):
        rematch = _anglere.match(line)
        if not rematch:
            raise ParameterError('Could not understand ANGLE line [%s]' % line)
        a1, a2, a3, k, eq = rematch.groups()
        a1 = a1.strip(); a2 = a2.strip(); a3 = a3.strip()
        typ = AngleType(float(k), float(eq))
        self.angle_types[(a1, a2, a3)] = typ
        self.angle_types[(a3, a2, a1)] = typ

    def _process_dihedral_line(self, line, finished_diheds, last_key):
        """ Processes a dihedral line, possibly part of a multi-term dihedral

        Parameters
        ----------
        line : str
            Line of the file that contains a dihedral term
        finished_diheds : dict
            Dictionary of dihedral parameters whose final term has been read in
            already (which means additional terms will overwrite, not add)
        last_key : str or None
            If not None, this is the key for the last dihedral type that should
            be implied if the atom types are missing. Atom types seem to only be
            required for the first term in a multi-term torsion definition

        Returns
        -------
        key or None
            If a negative periodicity indicates another term is coming, the
            current key is returned so it can be passed as key to the next
            _process_dihedral_call
        """
        rematch = _dihedre.match(line)
        if not rematch and last_key is None:
            raise ParameterError('Could not understand DIHEDRAL line '
                                 '[%s]' % line)
        elif not rematch:
            rematch = _dihed2re.match(line)
            if not rematch:
                raise ParameterError('Could not understand DIHEDRAL line '
                                     '[%s]' % line)
            div, k, phi, per = rematch.groups()
            key = last_key
            rkey = tuple(reversed(key))
            assert key in finished_diheds
            if finished_diheds[key]:
                raise AssertionError('Cannot have an implied torsion that '
                                     'has already finished!')
        else:
            a1, a2, a3, a4, div, k, phi, per = rematch.groups()
            a1, a2, a3, a4 =  a1.strip(), a2.strip(), a3.strip(), a4.strip()
            key = (a1, a2, a3, a4)
            rkey = (a4, a3, a2, a1)
            if last_key is not None and (last_key != key and last_key != rkey):
                warnings.warn('Expecting next term in dihedral %r, got '
                              'definition for dihedral %r' % (last_key, key),
                              ParameterWarning)
        scee = [float(x) for x in _sceere.findall(line)] or [1.2]
        scnb = [float(x) for x in _scnbre.findall(line)] or [2.0]
        per = float(per)
        typ = DihedralType(float(k)/float(div), abs(per), float(phi),
                           scee[0], scnb[0])
        if finished_diheds[key]:
            # This dihedral is already finished its definition, which means
            # we go ahead and add a new one to override it
            typs = DihedralTypeList()
            typs.append(typ)
            self.dihedral_types[key] = self.dihedral_types[rkey] = typs
        else:
            self.dihedral_types[key].append(typ)
        finished_diheds[key] = finished_diheds[rkey] = per >= 0
        if per < 0:
            return key

    def _process_improper_line(self, line):
        rematch = _impropre.match(line)
        if not rematch:
            raise ParameterError('Could not understand IMPROPER line '
                                    '[%s]' % line)
        a1, a2, a3, a4, k, phi, per = rematch.groups()
        a1 = a1.strip(); a2 = a2.strip();
        a3 = a3.strip(); a4 = a4.strip()
        # Pre-sort the improper types, assuming atom3 is the central atom (which
        # it must be in Amber parameter files!!!!)
        a1, a2, a4 = sorted([a1, a2, a4])
        key = (a1, a2, a3, a4)
        self.improper_periodic_types[key] = \
                DihedralType(float(k), float(per), float(phi))

    def _process_nonbond_line(self, line):
        try:
            atyp, rmin, eps = line.split()[:3]
        except ValueError:
            raise ParameterError('Could not understand nonbond parameter line '
                                 '[%s]' % line)
        try:
            self.atom_types[atyp].set_lj_params(float(eps), float(rmin))
        except KeyError:
            raise ParameterError('Atom type %s not present in the database.' %
                                 atyp)
        except ValueError:
            raise ParameterError('Could not convert nonbond parameters to '
                                 'floats [%s, %s]' % (rmin, eps))

    def _process_nbfix_line(self, line, equivalents=None):
        try:
            a1, a2, rmin1, eps1, rmin2, eps2 = line.split()[:6]
        except ValueError:
            raise ParameterError('Could not understand LJEDIT line [%s]' % line)
        try:
            rmin1 = float(rmin1)
            eps1 = float(eps1)
            rmin2 = float(rmin2)
            eps2 = float(eps2)
        except ValueError:
            raise ParameterError('Could not convert LJEDIT parameters '
                                 'to floats.')
        self.nbfix_types[(min(a1, a2), max(a1, a2))] = \
                (math.sqrt(eps1*eps2), rmin1+rmin2)
        if equivalents is not None:
            # We need to add the same nbfixes to all atom types that are
            # equivalent to the atom types defined in the LJEDIT line
            for oa1 in equivalents[a1]:
                self.nbfix_types[(min(oa1, a2), max(oa1, a2))] = \
                        (math.sqrt(eps1*eps2), rmin1+rmin2)
            for oa2 in equivalents[a2]:
                self.nbfix_types[(min(a1, oa2), max(a1, oa2))] = \
                        (math.sqrt(eps1*eps2), rmin1+rmin2)
            # Now do the equivalenced of atom 1 with the equivalenced of atom 2
            for oa1 in equivalents[a1]:
                for oa2 in equivalents[a2]:
                    self.nbfix_types[(min(oa1, oa2), max(oa1, oa2))] = \
                            (math.sqrt(eps1*eps2), rmin1+rmin2)

    #===================================================

    def write(self, dest, title='Created by ParmEd', style='frcmod'):
        """ Writes a parm.dat file with the current parameters

        Parameters
        ----------
        dest : str or file-like
            The file name or file-like object to write the parameters to
        title : str, optional
            The title of the frcmod to write. Default is 'Created by ParmEd'
        style : str, optional
            If 'frcmod', the parameters are written in frcmod-format. If 'parm',
            the parameters are written in parm.dat-format. Default is 'frcmod'
        """
        if isinstance(dest, string_types):
            outfile = genopen(dest, 'w')
            own_handle = True
        else:
            outfile = dest
            own_handle = False

        if style not in ('frcmod', 'parm'):
            raise ValueError('style must be either frcmod or parm, not %s' % style)

        outfile.write(title.rstrip('\r\n'))
        outfile.write('\n')
        # Write the atom mass
        outfile.write('MASS\n')
        for atom, typ in iteritems(self.atom_types):
            outfile.write('%s%6.3f\n' % (atom.ljust(6), typ.mass))
        outfile.write('\n')
        # Write the bonds
        outfile.write('BOND\n')
        done = set()
        for (a1, a2), typ in iteritems(self.bond_types):
            if id(typ) in done: continue
            done.add(id(typ))
            outfile.write('%s-%s   %8.3f  %6.3f\n' % (a1.ljust(2), a2.ljust(2), typ.k, typ.req))
        outfile.write('\n')
        # Write the angles
        outfile.write('ANGLE\n')
        done = set()
        for (a1, a2, a3), typ in iteritems(self.angle_types):
            if id(typ) in done: continue
            done.add(id(typ))
            outfile.write('%s-%s-%s   %8.3f  %6.3f\n' % (a1.ljust(2), a2.ljust(2), a3.ljust(2),
                                                         typ.k, typ.theteq))
        outfile.write('\n')
        # Write the dihedrals
        outfile.write('DIHE\n')
        done = set()
        for (a1, a2, a3, a4), typ in iteritems(self.dihedral_types):
            if id(typ) in done: continue
            done.add(id(typ))
            if isinstance(typ, DihedralType) or len(typ) == 1:
                if not isinstance(typ, DihedralType):
                    typ = typ[0]
                outfile.write('%s-%s-%s-%s %4i %14.8f %8.3f %5.1f    SCEE=%s SCNB=%s\n' %
                              (a1.ljust(2), a2.ljust(2), a3.ljust(2), a4.ljust(2), 1, typ.phi_k,
                               typ.phase, typ.per, typ.scee, typ.scnb))
            else:
                for dtyp in typ[:-1]:
                    outfile.write('%s-%s-%s-%s %4i %14.8f %8.3f %5.1f    SCEE=%s SCNB=%s\n' %
                                  (a1.ljust(2), a2.ljust(2), a3.ljust(2), a4.ljust(2), 1,
                                   dtyp.phi_k, dtyp.phase, -dtyp.per, dtyp.scee, dtyp.scnb))
                dtyp = typ[-1]
                outfile.write('%s-%s-%s-%s %4i %14.8f %8.3f %5.1f    SCEE=%s SCNB=%s\n' %
                              (a1.ljust(2), a2.ljust(2), a3.ljust(2), a4.ljust(2), 1, dtyp.phi_k,
                               dtyp.phase, dtyp.per, dtyp.scee, dtyp.scnb))
        outfile.write('\n')
        # Write the impropers
        outfile.write('IMPROPER\n')
        written_impropers = dict()
        for (a1, a2, a3, a4), typ in iteritems(self.improper_periodic_types):
            # Make sure wild-cards come at the beginning
            if a2 == 'X':
                assert a4 == 'X', 'Malformed generic improper!'
                a1, a2, a3, a4 = a2, a4, a3, a1
            elif a4 == 'X':
                a1, a2, a3, a4 = a4, a1, a3, a2
            a1, a2, a4 = sorted([a1, a2, a4])
            if (a1, a2, a3, a4) in written_impropers:
                if written_impropers[(a1, a2, a3, a4)] != typ:
                    raise ValueError('Multiple impropers with the same atom set not allowed')
                continue
            outfile.write('%s-%s-%s-%s %14.8f %8.3f %5.1f\n' %
                          (a1.ljust(2), a2.ljust(2), a3.ljust(2), a4.ljust(2), typ.phi_k,
                           typ.phase, typ.per))
            written_impropers[(a1, a2, a3, a4)] = typ
        outfile.write('\n')
        # Write the LJ terms
        outfile.write('NONB\n')
        for atom, typ in iteritems(self.atom_types):
            outfile.write('%s  %12.8f %12.8f\n' % (atom.ljust(2), typ.rmin, typ.epsilon))
        outfile.write('\n')
        # Write the NBFIX terms
        if self.nbfix_types:
            outfile.write('LJEDIT\n')
            for (a1, a2), (eps, rmin) in iteritems(self.nbfix_types):
                outfile.write('%s %s %12.8f %12.8f %12.8f %12.8f\n' %
                              (a1.ljust(2), a2.ljust(2), eps, rmin/2, eps, rmin/2))

        if own_handle:
            outfile.close()



"""
Contains the (very long) routine that does in-depth sanity checks on the
topology file
"""

from parmed.tools.exceptions import (AmberIncompatibleWarning, BadParmWarning,
        FixableParmWarning, NonfatalWarning, NonUniversalWarning,
        MissingDisulfide, LongBondWarning)
from parmed.constants import TINY
from parmed.amber.mask import AmberMask
from parmed.exceptions import MoleculeError
from parmed.utils.six.moves import range
from math import sqrt

def check_validity(parm, warnings):

    # local shortcut for _check_exist_nvals
    def checkme(*args, **kwargs):
        kwargs['warnings'] = warnings
        return _check_exist_nvals(parm, *args, **kwargs)

    # Make sure all of our sections that we expect are present and have the
    # right number of variables in it
    nttyp = parm.ptr('ntypes') * (parm.ptr('ntypes') + 1) // 2 # for LJ stuff
    checkme('ATOM_NAME', parm.ptr('natom'), True, False, None, str)
    checkme('CHARGE', parm.ptr('natom'), True, False, None, float)
    checkme('MASS', parm.ptr('natom'), True, False, None, float)
    checkme('ATOM_TYPE_INDEX', parm.ptr('natom'), True, False, None, int)
    checkme('NUMBER_EXCLUDED_ATOMS', parm.ptr('natom'), True, False, None, int)
    checkme('NONBONDED_PARM_INDEX', parm.ptr('ntypes')**2, True, False,
            None, int)
    checkme('RESIDUE_LABEL', parm.ptr('nres'), True, False, None, str)
    checkme('RESIDUE_POINTER', parm.ptr('nres'), True, False, None, int)
    checkme('BOND_FORCE_CONSTANT', parm.ptr('numbnd'), True, False, None, float)
    checkme('BOND_EQUIL_VALUE', parm.ptr('numbnd'), True, False, None, float)
    checkme('ANGLE_FORCE_CONSTANT', parm.ptr('numang'), True, False,
            None, float)
    checkme('ANGLE_EQUIL_VALUE', parm.ptr('numang'), True, False, None, float)
    checkme('DIHEDRAL_FORCE_CONSTANT', parm.ptr('nptra'), True,
            False, None, float)
    checkme('DIHEDRAL_PERIODICITY', parm.ptr('nptra'), True, False, None, float)
    checkme('DIHEDRAL_PHASE', parm.ptr('nptra'), True, False, None, float)
    checkme('SCEE_SCALE_FACTOR', parm.ptr('nptra'), False,
            True, 'scee 1.2', float)
    checkme('SCNB_SCALE_FACTOR', parm.ptr('nptra'), False,
            True, 'scnb 2.0', float)
    checkme('SOLTY', parm.ptr('natyp'), True, False, None, float)
    checkme('LENNARD_JONES_ACOEF', nttyp, True, False, None, float)
    checkme('LENNARD_JONES_BCOEF', nttyp, True, False, None, float)
    checkme('BONDS_INC_HYDROGEN', parm.ptr('nbonh')*3, True, False, None, int)
    checkme('BONDS_WITHOUT_HYDROGEN', parm.ptr('nbona')*3, True,
            False, None, int)
    checkme('ANGLES_INC_HYDROGEN', parm.ptr('ntheth')*4, True, False, None, int)
    checkme('ANGLES_WITHOUT_HYDROGEN', parm.ptr('ntheta')*4, True,
            False, None, int)
    checkme('DIHEDRALS_INC_HYDROGEN', parm.ptr('nphih')*5, True,
            False, None, int)
    checkme('DIHEDRALS_WITHOUT_HYDROGEN', parm.ptr('nphia')*5, True,
            False, None, int)
    checkme('EXCLUDED_ATOMS_LIST', parm.ptr('next'), True, False, None, int)
    checkme('HBOND_ACOEF', parm.ptr('nphb'), True, False, None, float)
    checkme('HBOND_BCOEF', parm.ptr('nphb'), True, False, None, float)
    checkme('HBCUT', parm.ptr('nphb'), True, False, None, float)
    checkme('AMBER_ATOM_TYPE', parm.ptr('natom'), True, False, None, str)
    checkme('TREE_CHAIN_CLASSIFICATION', parm.ptr('natom'), True,
            False, None, str)
    checkme('JOIN_ARRAY', parm.ptr('natom'), True, False, None, int)
    checkme('IROTAT', parm.ptr('natom'), True, False, None, int)
    # PBC tests
    if parm.ptr('ifbox') > 0:
        if checkme('SOLVENT_POINTERS', 3, True, True, 'setMolecules', int):
            checkme('ATOMS_PER_MOLECULE', parm.parm_data['SOLVENT_POINTERS'][1],
                    True, True, 'setMolecules', int)
        checkme('BOX_DIMENSIONS', 4, True, False, None, float)
        # Check for contiguous molecules. Back up ATOMS_PER_MOLECULE and
        # SOLVENT_POINTERS so we can use rediscover_molecules without clobbering
        # any changes users may have made
        apm_cpy = parm.parm_data['ATOMS_PER_MOLECULE'][:]
        sp_cpy = parm.parm_data['SOLVENT_POINTERS'][:]
        try:
            parm.rediscover_molecules(fix_broken=False)
            parm.parm_data['ATOMS_PER_MOLECULE'] = apm_cpy
            parm.parm_data['SOLVENT_POINTERS'] = sp_cpy
        except MoleculeError:
            warnings.warn('ATOMS_PER_MOLECULE section corrupt! Molecules are '
                          'not contiguous!')
    # Cap info
    if parm.ptr('ifcap') > 0:
        checkme('CAP_INFO', 1, True, False, None, int)
        checkme('CAP_INFO2', 4, True, False, None, float)
    # GB stuff
    if parm.ptr('ifbox') == 0:
        checkme('RADII', parm.ptr('natom'), True, True, 'changeRadii', float)
        checkme('SCREEN', parm.ptr('natom'), True, True, 'changeRadii', float)
        checkme('RADIUS_SET', 1, True, True, 'changeRadii', str)
        checkme('ATOMIC_NUMBER', parm.ptr('natom'),False,
                True, 'addAtomicNumber', int)
    # Some chamber checks
    if parm.chamber:
        # See if we have any CMAP terms
        has_cmap = 'CHARMM_CMAP_COUNT' in parm.flag_list
        # Ugh, don't want short-circuiting, so have to do this...
        try:
            nvals = parm.parm_data['FORCE_FIELD_TYPE'][0]
            nvals = int(nvals)
        except IndexError:
            warnings.warn('%FLAG FORCE_FIELD_TYPE is empty', BadParmWarning)
        except ValueError:
            warnings.warn('%FLAG FORCE_FIELD_TYPE does not have integer first',
                          BadParmWarning)
        if len(parm.parm_data['FORCE_FIELD_TYPE']) - 1 < nvals:
            warnings.warn('Not enough lines in %FLAG FORCE_FIELD_TYPE',
                          BadParmWarning)
        else:
            for i, val in enumerate(parm.parm_data['FORCE_FIELD_TYPE']):
                if i % 2 == 0 and not isinstance(val, int):
                    warnings.warn('int type mismatch in FORCE_FIELD_TYPE',
                                  BadParmWarning)
                elif i % 2 == 1 and not isinstance(val, str):
                    warnings.warn('str type mismatch in FORCE_FIELD_TYPE',
                                  BadParmWarning)
        hasallkeys = checkme('CHARMM_UREY_BRADLEY_COUNT', 2, False, False, None, int)
        hk = checkme('CHARMM_NUM_IMPROPERS', 1, True, False, None, int)
        hasallkeys = hasallkeys and hk
        hk = checkme('CHARMM_NUM_IMPR_TYPES', 1, True, False, None, int)
        hasallkeys = hasallkeys and hk
        if has_cmap:
            hk = checkme('CHARMM_CMAP_COUNT', 2, True, False, None, int)
            hasallkeys = hasallkeys and hk
     
        if hasallkeys:
            # Get the number of terms
            nub = parm.parm_data['CHARMM_UREY_BRADLEY_COUNT'][0]
            nubtypes = parm.parm_data['CHARMM_UREY_BRADLEY_COUNT'][1]
            nimphi = parm.parm_data['CHARMM_NUM_IMPROPERS'][0]
            nimprtyp = parm.parm_data['CHARMM_NUM_IMPR_TYPES'][0]
            cmap_term_cnt = parm.parm_data['CHARMM_CMAP_COUNT'][0]
            cmap_type_cnt = parm.parm_data['CHARMM_CMAP_COUNT'][1]
            # Look for terms
            checkme('LENNARD_JONES_14_ACOEF', nttyp, True, False, None, float)
            checkme('LENNARD_JONES_14_BCOEF', nttyp, True, False, None, float)
            checkme('CHARMM_UREY_BRADLEY', nub, True, False, None, int)
            checkme('CHARMM_UREY_BRADLEY_FORCE_CONSTANT', nubtypes, 
                    True, False, None, float)
            checkme('CHARMM_UREY_BRADLEY_EQUIL_CONSTANT', nubtypes,
                    True, False, None, float)
            checkme('CHARMM_IMPROPERS', nimphi*5, True, False, None, int)
            checkme('CHARMM_IMPROPER_FORCE_CONSTANT', nimprtyp, True,
                    False, None, float)
            checkme('CHARMM_IMPROPER_PHASE', nimprtyp, True, False, None, float)
            if has_cmap:
                checkme('CHARMM_CMAP_INDEX', cmap_term_cnt*6, True,
                        False, None,int)
                if checkme('CHARMM_CMAP_RESOLUTION', cmap_type_cnt, True,
                     False, True, int):
                    for i in range(cmap_type_cnt):
                        checkme('CHARMM_CMAP_PARAMETER_%s' % (str(i).zfill(2)),
                                parm.parm_data['CHARMM_CMAP_RESOLUTION'][i],
                                True, False, True, float)

    # Check that ATOMS_PER_MOLECULE == NATOM
    if 'ATOMS_PER_MOLECULE' in parm.parm_data:
        if sum(parm.parm_data['ATOMS_PER_MOLECULE']) != parm.ptr('natom'):
            warnings.warn('sum(ATOMS_PER_MOLECULE) != NATOM. Use '
                          '"setMolecules" to fix this', FixableParmWarning)

    # Duplicate pmemd's checks
    if parm.ptr('ifpert') != 0:
        warnings.warn('IFPERT must be 0! Parm will not work in Amber',
                        AmberIncompatibleWarning)
    if (parm.ptr('mbona') != parm.ptr('nbona') or
                parm.ptr('mtheta') != parm.ptr('ntheta') or
                parm.ptr('mphia') != parm.ptr('nphia')):
        warnings.warn('Constraints can no longer be put in the prmtop '
                      'in Amber programs!', AmberIncompatibleWarning)

    # Check that we didn't change off-diagonal LJ pairs, since this will not
    # work for all programs necessarily...  Check relative error, though, since
    # Lennard Jones coefficients are very large
    parm.fill_LJ()
    ntypes = parm.ptr('ntypes')
    try:
        for i in range(ntypes):
            for j in range(ntypes):
                idx = parm.parm_data['NONBONDED_PARM_INDEX'][ntypes*i+j] - 1
                rij = parm.LJ_radius[i] + parm.LJ_radius[j]
                wdij = sqrt(parm.LJ_depth[i] * parm.LJ_depth[j])
                acoef = parm.parm_data['LENNARD_JONES_ACOEF'][idx]
                bcoef = parm.parm_data['LENNARD_JONES_BCOEF'][idx]
                if acoef == 0 or bcoef == 0:
                    if acoef != 0 or bcoef != 0 or (wdij != 0 and rij != 0):
                        warnings.warn('Modified off-diagonal LJ parameters '
                                      'detected', NonUniversalWarning)
                        raise StopIteration
                elif (abs((acoef - (wdij * rij**12)) / acoef) > TINY or
                      abs((bcoef - (2 * wdij * rij**6)) / bcoef) > TINY):
                    warnings.warn('Modified off-diagonal LJ parameters '
                                  'detected', NonUniversalWarning)
                    raise StopIteration
    except StopIteration:
        pass

    # Check to see if we forgot to add disulfides. Unless we have coordinates,
    # though, we can only check that CYX sulfurs are bonded to another Sulfur
    mask = AmberMask(parm, ':CYX@SG')
    for i, sel in enumerate(mask.Selection()):
        if not sel: continue
        atm = parm.atoms[i]
        # We expect 2 bonds
        bondedatms = [a.name for a in atm.bond_partners]
        if len(bondedatms) != 2 or 'SG' not in bondedatms:
            warnings.warn('Detected CYX residue with a Sulfur atom not bonded '
                          'to another CYX Sulfur! Did you forget to make the '
                          'disulfide bond?', MissingDisulfide)
            break

    if parm.coordinates is not None:
        # Check if we think any disulfide bonds might be missing.
        mask = AmberMask(parm, ':CYS,CYM@SG')
        s_atms = []
        for i, sel in enumerate(mask.Selection()):
            if not sel: continue
            s_atms.append(parm.atoms[i])
        try:
            for i in range(len(s_atms)-1):
                for j in range(i, len(s_atms)):
                    dx = s_atms[i].xx - s_atms[j].xx
                    dy = s_atms[i].xy - s_atms[j].xy
                    dz = s_atms[i].xz - s_atms[j].xz
                    if (dx * dx + dy * dy + dz * dz) < 9.0:
                        warnings.warn(
                                "Detected two cysteine residues whose sulfur "
                                "atoms are within 3 Angstroms. Rename CYS to "
                                "CYX in the PDB file and use the 'bond' "
                                "command in tleap to create the disulfide "
                                "bond", MissingDisulfide
                        )
                        raise StopIteration
        except StopIteration:
            pass
        # Now check if we have any bonds that seem unreasonably large. This
        # would indicate gaps in the structure.
        for bnd in parm.bonds:
            dx = bnd.atom1.xx - bnd.atom2.xx
            dy = bnd.atom1.xy - bnd.atom2.xy
            dz = bnd.atom1.xz - bnd.atom2.xz
            d2 = dx*dx + dy*dy + dz*dz
            req = bnd.type.req
            # Warn if any bond starts at > 3 times its equilibrium length
            if d2 > 9 * req*req:
                warnings.warn('Atoms %d (%s %d [%s]) and %d (%s %d [%s]) are '
                    'bonded (equilibrium length %.3f A) but are %.3f A apart. '
                    'This often indicates gaps in the original sequence and '
                    'should be checked carefully.' %
                     (bnd.atom1.idx+1, bnd.atom1.residue.name,
                      bnd.atom1.residue.idx+1, bnd.atom1.name,
                      bnd.atom2.idx+1, bnd.atom2.residue.name,
                      bnd.atom2.residue.idx+1, bnd.atom2.name,
                      req, sqrt(d2)),
                    LongBondWarning
                )

def _check_exist_nvals(parm, key, nvals, required=True, addable=False,
                       addaction=None, typ=str, warnings=None):
    """
    Checks if a prmtop has a given key with the given number of values. We
    issue warnings based on whether it's a required flag and print advice
    if parmed can add it (i.e., addable)
    """
    if warnings is None:
        import warnings

    if addable and addaction is None:
        warnings.warn('Implementation Error: addable/addaction must be '
                        'specified together!')
        addable = False

    if not key in parm.parm_data:
        if not required:
            if addable:
                warnings.warn('%%FLAG %s not found, but it is not required. '
                              'It can be added to the prmtop using the '
                              '[[ %s ]] action in ParmEd' % (key, addaction),
                              FixableParmWarning)
            else:
                warnings.warn('%%FLAG %s not found, but it is not required.',
                            NonfatalWarning)
        else:
            if addable:
                warnings.warn('%%FLAG %s not found! Use the [[ %s ]] command '
                              'in ParmEd to add it.' % (key, addaction),
                              FixableParmWarning)
            else:
                warnings.warn('%%FLAG %s not found!' % key, BadParmWarning)
        # Bail out here or we get a KeyError
        return

    if len(parm.parm_data[key]) != nvals:
        warnings.warn('%%FLAG %s has %d values, expected %d!' % (key, 
                        len(parm.parm_data[key]), nvals))

    for val in parm.parm_data[key]:
        if not isinstance(val, typ):
            warnings.warn('Unexpected type for %%FLAG %s: expected %s, but '
                          'got %s' % (key, typ.__name__, type(val).__name__))
            break



from ._exceptions import *

def get_edit_descriptor_obj(name):
    '''Returns a new object instance from a string'''
    name = name.upper()
    if name == 'A':
        return A()
    elif name == 'B':
        return B()
    elif name == 'BN':
        return BN()
    elif name == 'BZ':
        return BZ()
    elif name == ':':
        return Colon()
    elif name == 'D':
        return D()
    elif name == 'E':
        return E()
    elif name == 'EN':
        return EN()
    elif name == 'ES':
        return ES()
    elif name == 'F':
        return F()
    elif name == 'G':
        return G()
    elif name == 'H':
        return H()
    elif name == 'I':
        return I()
    elif name == 'L':
        return L()
    elif name == 'O':
        return O()
    elif name == 'P':
        return P()
    elif name =='S':
        return S()
    elif name == '/':
        return Slash()
    elif name == 'SP':
        return SP()
    elif name == 'SS':
        return SS()
    elif name == 'T':
        return T()
    elif name == 'TL':
        return TL()
    elif name == 'TR':
        return TR()
    elif name == 'X':
        return X()
    elif name == 'Z':
        return Z()
    else:
        raise InvalidFormat('Expected an edit descriptor, got %s' % name)

# All the tokens defined in the F77 specification unless specified

class A(object):
    def __init__(self):
        self.repeat = None
        self.width = None
    def __repr__(self):
        return '<A repeat=' + str(self.repeat) + \
                ' width=' + str(self.width) + '>'

class QuotedString(object):
    def __init__(self, char_string=None):
        self.char_string = char_string
    def get_width(self):
        return len(self.char_string)
    width = property(get_width)
    def __repr__(self):
        return '<QuotedString char_string=' + str(self.char_string) + '>'

# Only in F95
class B(object):
    def __init__(self):
        self.repeat = None
        self.width = None
        self.min_digits = None
    def __repr__(self):
        return '<B repeat=' + str(self.repeat) + \
                ' width=' + str(self.width) + \
                ' min_digits=' + str(self.min_digits) + '>'
    
class BN(object):
    def __init__(self):
        pass
    def __repr__(self):
        return '<BN>'

class BZ(object):
    def __init__(self):
        pass
    def __repr__(self):
        return '<BZ>'

class Colon(object):
    def __init__(self):
        pass
    def __repr__(self):
        return '<Colon>'
    
class D(object):
    def __init__(self):
        self.repeat = None
        self.width = None
        self.decimal_places = None
    def __repr__(self):
        return '<D repeat=' + str(self.repeat) + \
                ' width=' + str(self.width) + \
                ' decimal_places=' + str(self.decimal_places) + '>'

class E(object):
    def __init__(self):
        self.repeat = None
        self.width = None
        self.decimal_places = None
        self.exponent = None
    def __repr__(self):
        return '<E repeat=' + str(self.repeat) + \
                ' width=' + str(self.width) + \
                ' decimal_places=' + str(self.decimal_places) + \
                ' exponent=' + str(self.exponent) + '>'
    
# Only in F95
class EN(object):
    def __init__(self):
        self.repeat = None
        self.width = None
        self.decimal_places = None
        self.exponent = None
    def __repr__(self):
        return '<EN repeat=' + str(self.repeat) + \
                ' width=' + str(self.width) + \
                ' decimal_places=' + str(self.decimal_places) + \
                ' exponent=' + str(self.exponent) + '>'

# Only in F95
class ES(object):
    def __init__(self):
        self.repeat = None
        self.width = None
        self.decimal_places = None
        self.exponent = None
    def __repr__(self):
        return '<ES repeat=' + str(self.repeat) + \
                ' width=' + str(self.width) + \
                ' decimal_places=' + str(self.decimal_places) + \
                ' exponent=' + str(self.exponent) + '>'

class F(object):
    def __init__(self):
        self.repeat = None
        self.width = None
        self.decimal_places = None
    def __repr__(self):
        return '<F repeat=' + str(self.repeat) + \
                ' width=' + str(self.width) + \
                ' decimal_places=' + str(self.decimal_places) + '>'
    
class FormatGroup(object):
    pass

class G(object):
    def __init__(self):
        self.repeat = None
        self.width = None
        self.decimal_places = None
        self.exponent = None
    def __repr__(self):
        return '<G repeat=' + str(self.repeat) + \
                ' width=' + str(self.width) + \
                ' decimal_places=' + str(self.decimal_places) + \
                ' exponent=' + str(self.exponent) + '>'

# Only in F77
class H(object):
    def __init__(self):
        self.num_chars = None
        self.char_string = None
    def __repr__(self):
        return '<H num_chars=' + str(self.num_chars) + \
                ' char_string=' + str(self.char_string) + '>'
    
class I(object):
    def __init__(self):
        self.repeat = None
        self.width = None
        self.min_digits = None
    def __repr__(self):
        return '<I repeat=' + str(self.repeat) + \
                ' width=' + str(self.width) + \
                ' min_digits=' + str(self.min_digits) + '>'
    
class L(object):
    def __init__(self):
        self.repeat = None
        self.width = None
    def __repr__(self):
        return '<L repeat=' + str(self.repeat) + \
                ' width=' + str(self.width) + '>'

# Only in F95
class O(object):
    def __init__(self):
        self.repeat = None
        self.width = None
        self.min_digits = None
    def __repr__(self):
        return '<O repeat=' + str(self.repeat) + \
                ' width=' + str(self.width) + \
                ' min_digits=' + str(self.min_digits) + '>'

class P(object):
    def __init__(self):
        self.scale = None
    def __repr__(self):
        return '<P scale=' + str(self.scale) + '>'
    
class S(object):
    def __init__(self):
        pass
    def __repr__(self):
        return '<S>'
    
class Slash(object):
    def __init__(self):
        self.repeat = None
        pass
    def __repr__(self):
        return '<Slash repeat=' + str(self.repeat) + '>'
    
class SP(object):
    def __init__(self):
        pass
    def __repr__(self):
        return '<SP>'
    
class SS(object):
    def __init__(self):
        pass
    def __repr__(self):
        return '<SS>'
    
class T(object):
    def __init__(self):
        self.num_chars = None
    def __repr__(self):
        return '<T num_chars=' + str(self.num_chars) + '>'
    
class TL(object):
    def __init__(self):
        self.num_chars = None
    def __repr__(self):
        return '<TL num_chars=' + str(self.num_chars) + '>'
    
class TR(object):
    def __init__(self):
        self.num_chars = None
    def __repr__(self):
        return '<TR num_chars=' + str(self.num_chars) + '>'

class X(object):
    def __init__(self):
        self.num_chars = None
    def __repr__(self):
        return '<X num_chars=' + str(self.num_chars) + '>'

# Only in F95
class Z(object):
    def __init__(self):
        self.repeat = None
        self.width = None
        self.min_digits = None
    def __repr__(self):
        return '<Z repeat=' + str(self.repeat) + \
                ' width=' + str(self.width) + \
                ' min_digits=' + str(self.min_digits) + '>'

# Categorise the edit descriptors depnding on how they should be parsed

ED1 = ['BN', 'BZ', 'SP', 'SS', 'S'] # Of form X only
ED2 = ['X'] # Of form nX only
ED3 = ['T', 'TR', 'TL', 'L'] # Of form Xn only
ED4 = ['A'] # Of form X or Xn
ED5 = ['D', 'F'] # Of form Xn.m only
ED6 = ['B', 'I', 'O', 'Z'] # Of form Xn or Xn.m
ED7 = ['E', 'EN', 'ES', 'G'] # Of form Xn.m or Xn.mEe
ED8 = ['P'] # Of form kX only, where k is a signed integer, may omit comma if followed by Type 5 or 7 edit descriptor
ED9 = [':'] # Of form X only, may omit comma either side
ED10 = ['/'] # Of form X only, may omit following comma and leading comma if no repeat
REPEATABLE_EDS = ['L', 'A', 'D', 'F', 'B', 'I', 'O', 'Z', 'E', 'EN', 'ES', 'G', '/']
OUTPUT_EDS = (L, A, D, F, B, I, O, Z, E, EN, ES, G)
CONTROL_EDS = (BN, BZ, P, SP, SS, S, X, T, TR, TL, Colon, Slash)
NON_REVERSION_EDS = (P, S, SP, SS, BN, BZ)
ALL_ED = ED1 + ED2 + ED3 + ED4 + ED5 + ED6 + ED7 + ED8 + ED9 + ED10




import re
import pdb
from ._edit_descriptors import *
from ._misc import expand_edit_descriptors, has_next_iterator
from . import config

WIDTH_OPTIONAL_EDS = [A]
NON_WIDTH_EDS = [BN, BZ, P, SP, SS, S, X, T, TR, TL, Colon, Slash]
FORBIDDEN_EDS = [QuotedString, H]

# Some problems without pre written input vars:
#   Cannot say when reversion conditions are met
#   Cannot determine width of A edit descriptor
#   Cannot determine complex input
#   Cannot determine proper input for G edit descriptors


def input(eds, reversion_eds, records, num_vals=None):

    state = { \
        'position' : 0,
        'scale' : 0,
        'incl_plus' : False,
        'blanks_as_zeros' : config.PROC_BLANKS_AS_ZEROS,
        # TODO: Implement halt if no more record input
        'halt_if_no_vals' : False,
        'exception_on_fail' : True,
    }

    # pdb.set_trace()

    for ed in eds + reversion_eds:
        if isinstance(ed, tuple(FORBIDDEN_EDS)):
            raise InvalidFormat("%d edit descriptr not permitted on input")

    # Expand repeated edit decriptors
    eds = expand_edit_descriptors(eds)
    reversion_eds = expand_edit_descriptors(reversion_eds)
    # Assume one-to-one correspondance between edit descriptors and output
    # values if number of output values is not defined 
    num_out_eds = 0
    for ed in eds:
        if isinstance(ed, OUTPUT_EDS):
            num_out_eds += 1
    num_rev_out_eds = 0
    if num_vals is None:
        num_vals = num_out_eds
    for ed in reversion_eds:
        if isinstance(ed, OUTPUT_EDS):
            num_rev_out_eds += 1

    
    # Will loop forever is no output edit descriptors
    if (num_out_eds == 0):
        return []
    # Will loop forever if no output eds in reversion format and is more values
    # requested than in the format
    if (num_vals > num_out_eds) and (num_rev_out_eds == 0):
        raise ValueError('Not enough output edit descriptors in reversion format to output %d values' % num_vals)

    # May need to process multiple records, down to a higher function to supply
    # appropriate string for format
    if not hasattr(records, 'next'):
        records = iter(re.split('\r\n|\r|\n', records))
    record = _next(records, None)
    if record is None:
        return [] 
    
    # if a_widths is not None:
    #     a_widths = itertools.cycle(a_widths)

    vals = []
    finish_up = False
    ed_ind = -1
    while True:
        ed_ind += 1
        # Signal to stop when Colon edit descriptor or end of format or end of
        # reversion format reached. Also not to output any more data
        if len(vals) >= num_vals:
            finish_up = True
        # Select the appropriate edit descriptor
        if ed_ind < len(eds):
            ed = eds[ed_ind]
        else:
            rev_ed_ind = (ed_ind - len(eds)) % len(reversion_eds)
            # Reversion begun and has been instructed to halt
            if finish_up and (rev_ed_ind == 0):
                break
            ed = reversion_eds[rev_ed_ind]

        if isinstance(ed, QuotedString):
            raise InvalidFormat('Cannot have string literal in an input format')
        elif isinstance(ed, BN):
            state['blanks_as_zeros'] = False
        elif isinstance(ed, BZ):
            state['blanks_as_zeros'] = True
        elif isinstance(ed, P):
            state['scale'] = ed.scale
        elif isinstance(ed, SP):
            state['incl_plus'] = True
        elif isinstance(ed, SS):
            state['incl_plus'] = False
        elif isinstance(ed, S):
            state['incl_plus'] = config.PROC_INCL_PLUS
        elif isinstance(ed, (X, TR)):
            state['position'] = min(state['position'] + ed.num_chars, len(record))
        elif isinstance(ed, TL):
            state['position'] = max(state['position'] - ed.num_chars, 0)
        elif isinstance(ed, T):
            if (ed.num_chars - 1) < 0:
                state['position'] = 0
            elif ed.num_chars > len(record):
                state['position'] = len(record)
            else:
                state['position'] = ed.num_chars - 1
        elif isinstance(ed, Slash):
            # End of record
            record = _next(records, None)
            state['position'] = 0
            if record is None:
                break
        elif isinstance(ed, Colon):
            # Break if input value satisfied
            if finish_up:
                break
        elif isinstance(ed, (Z, O, B, I)):
            val, state = read_integer(ed, state, record)
            vals.append(val)
        elif isinstance(ed, A):
            val, state = read_string(ed, state, record)
            vals.append(val)
        elif isinstance(ed, L):
            val, state = read_logical(ed, state, record)
            vals.append(val)
        elif isinstance(ed, (F, E, D, EN, ES)):
            val, state = read_float(ed, state, record)
            vals.append(val)
        elif isinstance(ed, G):
            # Difficult to know what wanted since do not know type of input variable
            # Use the G_INPUT_TRIAL_EDS variable to try the variables
            # until one sticks
            # n.b. vals and state do not get written to if
            # exception id raised
            resolved = False
            g_trial_eds = iter(config.G_INPUT_TRIAL_EDS)
            while not resolved:
                ed_name = _next(g_trial_eds, '')
                if ed_name.upper() in ('F', 'E', 'D', 'EN', 'ES'):
                    trial_ed = F()
                    trial_ed.width = ed.width
                    trial_ed.decimal_places = ed.decimal_places
                    # pdb.set_trace()
                    try:
                        val, state = read_float(trial_ed, state.copy(), record)
                        vals.append(val)
                        resolved = True
                    except ValueError:
                        continue
                elif ed_name.upper() in ('Z', 'O', 'B', 'I'):
                    trial_ed = globals()[ed_name]()
                    trial_ed.width = ed.width
                    trial_ed.min_digits = ed.decimal_places
                    try:
                        val, state = read_integer(trial_ed, state.copy(), record)
                        vals.append(val)
                        resolved = True
                    except ValueError:
                        continue
                elif ed_name.upper() in ('L'):
                    trial_ed = L()
                    trial_ed.width = ed.width
                    try:
                        val, state = read_logical(trial_ed, state.copy(), record)
                        vals.append(val)
                        resolved = True
                    except ValueError:
                        continue
                elif ed_name.upper() in ('A'):
                    trial_ed = A()
                    trial_ed.width = ed.width
                    try:
                        val, state = read_string(trial_ed, state.copy(), record)
                        vals.append(val)
                        resolved = True
                    except ValueError:
                        continue
                elif ed_name in ('G'):
                    raise ValueError('G edit descriptor not permitted in config.G_INPUT_TRIAL_EDS')
                else:
                    raise ValueError('Unrecognised trial edit descriptor string in config.G_INPUT_TRIAL_EDS')
                    
    if config.RET_WRITTEN_VARS_ONLY:
        vals = [val for val in vals if val is not None]
    return vals[:num_vals]

def _interpret_blanks(substr, state):
    # Save leading blanks
    len_str = len(substr)
    if state['blanks_as_zeros']:
        # TODO: Are tabs blank characters?
        substr = substr.replace(' ', '0')
    else:
        substr = substr.replace(' ', '')
    # If were blanks but have been stripped away, replace with a zero
    if len(substr) == 0 and (len_str > 0):
        substr = '0'
    return substr

def _get_substr(w, record, state):
    start = max(state['position'], 0)
    end = start + w
    # if end > len(record):
    #     substr = ''
    #     # TODO: test if no chars transmitted, then poition does not change
    #     w = 0
    # else:
    substr = record[start:end]
    state['position'] = min(state['position'] + w, len(record))
    return substr, state


def _next(it, default=None):
    try:
        val = next(it)
    except StopIteration:
        val = default
    return val


def read_string(ed, state, record):
    if ed.width is None:
        # Will assume rest of record is fair game for the
        # unsized A edit descriptor
        ed.width = len(record) - state['position']
    substr, state = _get_substr(ed.width, record, state)
    val = substr.ljust(ed.width, config.PROC_PAD_CHAR)
    return (val, state)


def read_integer(ed, state, record):
    substr, state = _get_substr(ed.width, record, state)
    if ('-' in substr) and (not config.PROC_ALLOW_NEG_BOZ) and isinstance(ed, (Z, O, B)):
        if state['exception_on_fail']:
            raise ValueError('Negative numbers not permitted for binary, octal or hex')
        else:
            return (None, state)
    if isinstance(ed, Z):
        base = 16
    elif isinstance(ed, I):
        base = 10
    elif isinstance(ed, O):
        base = 8
    elif isinstance(ed, B):
        base = 2
    # If a negative is followed by blanks, Gfortran and ifort
    # interpret as a zero
    if re.match(r'^ *- +$', substr):
        substr = '0'
    # If a negative or negative and blanks, ifort interprets as
    # zero for an I edit descriptor
    if config.PROC_NEG_AS_ZERO and isinstance(ed, I) and re.match(r'^( *- *| +)$', substr):
        substr = '0'
    # If string is zero length (reading off end of record?),
    # interpret as zero so as to match what would be found in an
    # unwritten FORTRAN variable
    if substr == '':
        if config.RET_UNWRITTEN_VARS_NONE or config.RET_WRITTEN_VARS_ONLY:
            return (None, state)
        else:
            substr = '0'
    teststr = _interpret_blanks(substr, state)
    try:
        val = int(teststr, base)
    except ValueError:
        if state['exception_on_fail']:
            raise ValueError('%s is not a valid input for one of integer, octal, hex or binary' % substr)
        else:
            return (None, state)
    return (val, state)


def read_logical(ed, state, record):
    substr, state = _get_substr(ed.width, record, state)
    # Deal with case where there is no more input to read from
    if (substr == '') and (config.RET_UNWRITTEN_VARS_NONE or config.RET_WRITTEN_VARS_ONLY):
        return (None, state)
    # Remove preceding whitespace and take the first two letters as
    # uppercase for testing
    teststr = substr.upper().lstrip().lstrip('.')
    if len(teststr):
        teststr = teststr[0]
    else:
        # This is case where just a preceding period is read in
        raise ValueError('%s is not a valid boolean input' % substr)
    if teststr == 'T':
        val = True
    elif teststr == 'F':
        val = False
    else:
        if state['exception_on_fail']:
            raise ValueError('%s is not a valid boolean input' % substr)
        else:
            val = None
    return (val, state)


def read_float(ed, state, record):
    substr, state = _get_substr(ed.width, record, state)
    teststr = _interpret_blanks(substr, state)
    # When reading off end of record, get empty string,
    # interpret as 0
    if teststr == '':
        if config.RET_UNWRITTEN_VARS_NONE or config.RET_WRITTEN_VARS_ONLY:
            return (None, state)
        else:
            teststr = '0'
    # Python only understands 'E' as an exponential letter
    teststr = teststr.upper().replace('D', 'E')
    # Prepend an exponential letter if only a '-' or '+' denotes an exponent
    if 'E' not in teststr:
        teststr = teststr[0] + teststr[1:].replace('+', 'E+').replace('-', 'E-')
    # ifort allows '.' to be interpreted as 0
    if re.match(r'^ *\. *$', teststr):
        teststr = '0'
    # ifort allows '-' to be interpreted as 0
    if re.match(r'^ *- *$', teststr):
        teststr = '0'
    # ifort allows numbers to end with 'E', 'E+', 'E-' and 'D'
    # equivalents
    res = re.match(r'(.*)(E|E\+|E\-)$', teststr)
    if res:
        teststr = res.group(1)
    try:
        val = float(teststr)
    except ValueError:
        if state['exception_on_fail']:
            raise ValueError('%s is not a valid input as for an E, ES, EN or D edit descriptor' % substr)
        else:
            return (None, state)
    # Special cases: insert a decimal if none specified
    if ('.' not in teststr) and (ed.decimal_places is not None):
        val = val / 10 ** ed.decimal_places
    # Apply scale factor if exponent not supplied
    if 'E' not in teststr:
        val = val / 10 ** state['scale'] 
    return (val, state)



"""
The PDBx package here was taken from the wwPDB website written by John Westbook
(and modified slightly to fix issues with Python 3 support as well as unified
exception handling)
"""

__author__ = "John Westbrook"
__contributors__ = "Jason Swails"
__version__ = "V0.01"
__license__ = "Creative Commons Attribution 3.0 Unported"
__email__ = "jwest@rcsb.rutgers.edu <or> jason.swails@gmail.com"

__all__ = ['PdbxReader', 'PdbxWriter', 'containers']

from parmed.formats.pdbx.PdbxReader import PdbxReader
from parmed.formats.pdbx.PdbxWriter import PdbxWriter
from parmed.formats.pdbx import PdbxContainers as containers
